def test_getlogs(web3_eth_tester):
    """test that getLogs works with a list of addresses

    there is a regression in web3 4.5.0, which breaks our usage of getLogs see
    https://github.com/ethereum/web3.py/issues/999
    """
    web3_eth_tester.eth.getLogs(
        {
            "fromBlock": 1,
            "toBlock": 1,
            "address": [
                "0x55bdaAf9f941A5BB3EacC8D876eeFf90b90ddac9",
                "0xC0B33D88C704455075a0724AA167a286da778DDE",
            ],
        }
    )



import os
import json
import testing.postgresql
import pytest
import psycopg2
import psycopg2.extras
import eth_tester
from web3 import Web3
from web3.providers.eth_tester import EthereumTesterProvider
import attr
from typing import List, Any
from ethindex import logdecode, pgimport
import eth_utils


@pytest.fixture(scope="session")
def contracts_json_path():
    return os.path.normpath(os.path.join(__file__, "..", "build/contracts.json"))


@pytest.fixture(scope="session")
def compiled_contracts(contracts_json_path):
    with open(contracts_json_path) as contracts_json:
        return json.load(contracts_json)


@pytest.fixture(scope="session")
def postgresql_factory():
    # Generate Postgresql class which shares the generated database
    factory = testing.postgresql.PostgresqlFactory(cache_initialized_db=True)
    yield factory
    factory.clear_cache()


@pytest.fixture
def postgresql(postgresql_factory):
    postgresql = postgresql_factory()
    dsn = postgresql.dsn()
    os.environ["PGDATABASE"] = dsn["database"]
    os.environ["PGHOST"] = dsn["host"]
    os.environ["PGPORT"] = str(dsn["port"])
    os.environ["PGUSER"] = dsn["user"]
    yield postgresql
    postgresql.stop()


@pytest.fixture
def postgresql_dsn(postgresql):
    return postgresql.dsn()


@pytest.fixture
def conn(postgresql_dsn):
    return psycopg2.connect(
        **postgresql_dsn, cursor_factory=psycopg2.extras.RealDictCursor
    )


@pytest.fixture(scope="session")
def ethereum_tester():
    """Returns an instance of an Ethereum tester"""
    tester = eth_tester.EthereumTester(eth_tester.PyEVMBackend())
    return tester


def deploy_contract(web3, contract_interface):
    CurrencyNetwork = web3.eth.contract(
        abi=contract_interface["abi"], bytecode=contract_interface["bytecode"]
    )
    tx_hash = CurrencyNetwork.constructor().transact()
    tx_receipt = web3.eth.waitForTransactionReceipt(tx_hash)
    return tx_receipt.contractAddress


@attr.s(auto_attribs=True)
class TestEnv:
    web3: Web3
    contract_addresses: List[str]
    currency_networks: List[Any]
    addresses_json_path: str
    contracts_json_path: str
    contracts: List[Any]
    abi: Any
    topic_index: logdecode.TopicIndex
    ethereum_tester: eth_tester.EthereumTester


@pytest.fixture
def testenv(
    ethereum_tester: eth_tester.EthereumTester,
    compiled_contracts,
    tmpdir,
    contracts_json_path,
):
    web3 = Web3(EthereumTesterProvider(ethereum_tester))
    web3.eth.defaultAccount = web3.eth.accounts[0]

    contract_interface = compiled_contracts["CurrencyNetwork"]
    contract_addresses = [deploy_contract(web3, contract_interface) for i in range(3)]
    currency_networks = [
        web3.eth.contract(address=contract_address, abi=contract_interface["abi"])
        for contract_address in contract_addresses
    ]

    addresses_json_path = tmpdir.join("addresses.json")
    addresses_json_path.write(json.dumps({"networks": contract_addresses}))
    snapshot = ethereum_tester.take_snapshot()
    abi = contract_interface["abi"]
    yield TestEnv(
        web3,
        contract_addresses,
        currency_networks,
        str(addresses_json_path),
        contracts_json_path,
        [
            web3.eth.contract(address=contract_address, abi=contract_interface["abi"])
            for contract_address in contract_addresses
        ],
        abi,
        logdecode.TopicIndex(
            {contract_address: abi for contract_address in contract_addresses}
        ),
        ethereum_tester,
    )
    ethereum_tester.revert_to_snapshot(snapshot)


@pytest.fixture
def web3_eth_tester(testenv):
    return testenv.web3


def make_address(i: int):
    return eth_utils.to_checksum_address("0x{:040X}".format(i))


class EventEmitter:
    """this class is used to emit events into the blockchain"""

    def __init__(self, testenv):
        self.testenv = testenv
        self.value = 0

    def add_some_tranfer_events(self):
        for i, contract in enumerate(self.testenv.contracts):
            contract.functions.makeTransfer(
                make_address(i), make_address(i + 1), self.value
            ).transact()
            self.value += 1


@pytest.fixture
def event_emitter(testenv):
    return EventEmitter(testenv)


@pytest.fixture
def synchronizer(testenv, conn):
    pgimport.do_createtables(conn)
    pgimport.do_importabi(
        conn, testenv.addresses_json_path, testenv.contracts_json_path
    )
    pgimport.ensure_default_entry(conn)
    synchronizer = pgimport.Synchronizer(
        conn, testenv.web3, "default", required_confirmations=10
    )
    return synchronizer



"""test that chain reorgs are handled"""

import pytest
import logging


def fetch_events(conn):
    with conn.cursor() as cur:
        cur.execute("select * from events order by blocknumber")
        rows = cur.fetchall()
        return [event["args"]["_value"] for event in rows]


def test_reorg(testenv, event_emitter, conn, synchronizer):
    # event_emitter.add_some_tranfer_events adds 3 events and increases the value
    # added by one for each event

    event_emitter.add_some_tranfer_events()  # add events with values 0, 1, 2
    snapshot = testenv.ethereum_tester.take_snapshot()
    event_emitter.add_some_tranfer_events()  # add events with values 3, 4, 5
    synchronizer.sync_until_current()
    values_before = fetch_events(conn)
    assert values_before == [0, 1, 2, 3, 4, 5]

    # now reorg the chain
    testenv.ethereum_tester.revert_to_snapshot(snapshot)
    event_emitter.add_some_tranfer_events()  # add events with values 6, 7, 8
    event_emitter.add_some_tranfer_events()  # add events with values 9, 10, 11
    synchronizer.sync_until_current()
    values_after = fetch_events(conn)
    assert values_after == [0, 1, 2, 6, 7, 8, 9, 10, 11]


@pytest.mark.xfail
def test_reorg_shorter_chain(testenv, event_emitter, conn, synchronizer):
    """test that reorgs that result in a shorter chain are handled

    The chain could get shorter since it only has to be the longest chain difficulty-wise.
    Not sure if this is only a theoretical problem.
    """
    # event_emitter.add_some_tranfer_events adds 3 events and increases the value
    # added by one for each event

    event_emitter.add_some_tranfer_events()  # add events with values 0, 1, 2
    snapshot = testenv.ethereum_tester.take_snapshot()
    event_emitter.add_some_tranfer_events()  # add events with values 3, 4, 5
    event_emitter.add_some_tranfer_events()  # add events with values 6, 7, 8
    synchronizer.sync_until_current()
    values_before = fetch_events(conn)
    print("values_before", values_before)
    assert values_before == [0, 1, 2, 3, 4, 5, 6, 7, 8]

    # now reorg the chain
    testenv.ethereum_tester.revert_to_snapshot(snapshot)
    event_emitter.add_some_tranfer_events()  # add events with values 9, 10, 11
    synchronizer.sync_until_current()
    values_after = fetch_events(conn)
    print("values_after", values_after)
    assert values_after == [0, 1, 2, 9, 10, 11]


def test_sync_until_current(testenv, event_emitter, conn, synchronizer, caplog):
    # event_emitter.add_some_tranfer_events adds 3 events and increases the value
    # added by one for each event
    caplog.set_level(logging.INFO)
    synchronizer.blocks_per_round = 23
    synchronizer.required_confirmations = 0
    synchronizer.sync_until_current()

    for _ in range(11):
        event_emitter.add_some_tranfer_events()

    synchronizer.sync_until_current()

    values = fetch_events(conn)
    print(values)
    assert len(values) == 33



"""ethereum log decoding

This module provides some helper functions to do ethereum log decoding

See https://codeburst.io/deep-dive-into-ethereum-logs-a8d2047c7371 for a nice
explanation.
"""

import hexbytes
import eth_abi
import itertools
import eth_utils
import attr
from typing import List, Any, Dict, Optional


def replace_with_checksum_address(values: List[Any], types: List[str]) -> List[Any]:
    """returns a new list of values with addresses replaced with their checksum
    address """
    return [
        eth_utils.to_checksum_address(value) if _type == "address" else value
        for (value, _type) in zip(values, types)
    ]


def build_address_to_abi_dict(
    addresses_json: Dict[str, Any], compiled_contracts: Dict
) -> Dict[str, Dict]:
    """build a contract-address to abi mapping from addresses.json and the contracts.json

    This function doesn't read the json files, but instead expects the decoded
    json from those files.
    """
    res = {}

    def add_abi(contract_address, contract_name):
        res[eth_utils.to_checksum_address(contract_address)] = compiled_contracts[
            contract_name
        ]["abi"]

    for network in addresses_json["networks"]:
        add_abi(network, "CurrencyNetwork")

    if "unwEth" in addresses_json:
        add_abi(addresses_json["unwEth"], "UnwEth")
    if "exchange" in addresses_json:
        add_abi(addresses_json["exchange"], "Exchange")
    return res


def decode_non_indexed_inputs(abi, log):
    """decode non-indexed inputs from a log entry

    This one decodes the values stored in the 'data' field."""
    inputs = [input_ for input_ in abi["inputs"] if not input_["indexed"]]
    types = [input_["type"] for input_ in inputs]
    names = [input_["name"] for input_ in inputs]
    data = hexbytes.HexBytes(log["data"])
    values = eth_abi.decode_abi(types, data)
    return zip(names, replace_with_checksum_address(values, types))


def decode_indexed_inputs(abi, log):
    """decode indexed inputs from a log entry

    This one decodes the values stored in in topics fields (without topic 0)"""
    inputs = [input_ for input_ in abi["inputs"] if input_["indexed"]]
    types = [input_["type"] for input_ in inputs]
    names = [input_["name"] for input_ in inputs]
    values = [
        eth_abi.decode_single(type_, value)
        for type_, value in zip(types, log["topics"][1:])
    ]
    return zip(names, replace_with_checksum_address(values, types))


def get_event_abis(abi):
    return [some_abi for some_abi in abi if some_abi["type"] == "event"]


@attr.s(auto_attribs=True)
class Event:
    name: str
    args: Dict
    log: Dict
    timestamp: Optional[int]

    @property
    def blocknumber(self):
        return self.log["blockNumber"]

    @property
    def blockhash(self):
        return self.log["blockHash"]

    @property
    def transactionhash(self):
        return self.log["transactionHash"]

    @property
    def address(self):
        return self.log["address"]

    @property
    def transactionindex(self):
        return self.log["transactionIndex"]

    @property
    def logindex(self):
        return self.log["logIndex"]


class TopicIndex:
    def __init__(self, address2abi):
        """build a TopicIndex from an contract address to ABI dict"""
        self.addresses = list(address2abi.keys())
        self.address2abi = address2abi
        self.address_topic2event_abi = {}
        for address, abi in self.address2abi.items():
            for event_abi in get_event_abis(abi):
                self.address_topic2event_abi[
                    (
                        address,
                        hexbytes.HexBytes(eth_utils.event_abi_to_log_topic(event_abi)),
                    )
                ] = event_abi

    def get_abi_for_log(self, log):
        return self.address_topic2event_abi.get((log["address"], log["topics"][0]))

    def decode_log(self, log) -> Event:
        abi = self.get_abi_for_log(log)
        if abi is None:
            raise RuntimeError("Could not find ABI for log %s", log)

        args = dict(
            itertools.chain(
                decode_non_indexed_inputs(abi, log), decode_indexed_inputs(abi, log)
            )
        )
        return Event(name=abi["name"], args=args, log=log, timestamp=None)



from ethindex import pgimport


def test_get_events(testenv, event_emitter):
    event_emitter.add_some_tranfer_events()
    events = pgimport.get_events(testenv.web3, testenv.topic_index, 0, "latest")
    print("EVENTS:", len(events), events)

    assert [event.args for event in events] == [
        {
            "_value": 0,
            "_from": "0x0000000000000000000000000000000000000000",
            "_to": "0x0000000000000000000000000000000000000001",
        },
        {
            "_value": 1,
            "_from": "0x0000000000000000000000000000000000000001",
            "_to": "0x0000000000000000000000000000000000000002",
        },
        {
            "_value": 2,
            "_from": "0x0000000000000000000000000000000000000002",
            "_to": "0x0000000000000000000000000000000000000003",
        },
    ]


def test_topic_index_from_db(testenv, event_emitter, conn):
    event_emitter.add_some_tranfer_events()
    pgimport.do_createtables(conn)
    pgimport.do_importabi(
        conn, testenv.addresses_json_path, testenv.contracts_json_path
    )

    topic_index2 = pgimport.topic_index_from_db(conn)
    topic_index3 = pgimport.topic_index_from_db(conn, testenv.contract_addresses)

    assert set(topic_index2.addresses) == set(testenv.contract_addresses)
    assert set(topic_index3.addresses) == set(testenv.contract_addresses)

    events1 = pgimport.get_events(testenv.web3, testenv.topic_index, 0, "latest")
    assert events1
    events2 = pgimport.get_events(testenv.web3, topic_index2, 0, "latest")
    events3 = pgimport.get_events(testenv.web3, topic_index3, 0, "latest")
    assert events1 == events2
    assert events1 == events3



import click
import pkg_resources

import ethindex.pgimport


def report_version():
    click.echo(pkg_resources.get_distribution("eth-index").version)


@click.group(invoke_without_command=True)
@click.option("--version", help="Prints the version of the software", is_flag=True)
@click.pass_context
def cli(ctx, version):
    if version:
        report_version()
    elif ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()


cli.add_command(ethindex.pgimport.importabi)
cli.add_command(ethindex.pgimport.runsync)
cli.add_command(ethindex.pgimport.createtables)
cli.add_command(ethindex.pgimport.droptables)



import subprocess


def find_tables(conn):
    """return a list of tables"""
    with conn.cursor() as cur:
        cur.execute(
            "select table_name from information_schema.tables where table_schema='public'"
        )
        return cur.fetchall()


def test_createtables(conn):
    err = subprocess.call(["ethindex", "createtables"])
    assert err == 0
    tables = find_tables(conn)
    print("FOUND", tables)
    assert len(tables) >= 3


def test_droptables_without_force(conn):
    """droptables require the user to pass --force"""
    err = subprocess.call(["ethindex", "createtables"])
    assert err == 0
    tables_before = find_tables(conn)
    err = subprocess.call(["ethindex", "droptables"])
    assert err == 1
    tables_after = find_tables(conn)
    assert tables_before == tables_after


def test_droptables_with_force(conn):
    err = subprocess.call(["ethindex", "createtables"])
    assert err == 0
    err = subprocess.call(["ethindex", "droptables", "--force"])
    assert err == 0

    tables = find_tables(conn)
    assert tables == []


def test_importabi(conn, testenv):
    err = subprocess.call(["ethindex", "createtables"])
    assert err == 0
    err = subprocess.call(
        [
            "ethindex",
            "importabi",
            "--addresses",
            testenv.addresses_json_path,
            "--contracts",
            testenv.contracts_json_path,
        ]
    )
    assert err == 0
    with conn.cursor() as cur:
        cur.execute("select * from abis")
        abis = cur.fetchall()
        assert len(abis) == len(testenv.contract_addresses)



"""import ethereum events into postgres
"""
import sys
import json
import time
from web3 import Web3
import psycopg2
import psycopg2.extras
import binascii
import logging
import click
from ethindex import logdecode, util
from typing import Iterable

logger = logging.getLogger(__name__)
# https://github.com/ethereum/wiki/wiki/JavaScript-API#web3ethgettransactionreceipt


def topic_index_from_db(conn, addresses=None):
    """create a logdecode.TopicIndex from the ABIs stored in the abis table"""
    with conn.cursor() as cur:
        if addresses is None:
            cur.execute("select * from abis")
        else:
            cur.execute(
                "select * from abis where contract_address in %s", (tuple(addresses),)
            )
        rows = cur.fetchall()
        return logdecode.TopicIndex({r["contract_address"]: r["abi"] for r in rows})


def get_logs(web3, addresses, fromBlock, toBlock):
    fromBlock = hex(fromBlock)
    if toBlock != "latest":
        toBlock = hex(toBlock)

    return web3.eth.getLogs(
        {"fromBlock": fromBlock, "toBlock": toBlock, "address": addresses}
    )


def get_events(web3, topic_index, fromBlock, toBlock) -> Iterable[logdecode.Event]:
    return [
        topic_index.decode_log(x)
        for x in get_logs(web3, topic_index.addresses, fromBlock, toBlock)
    ]


def hexlify(d):
    return "0x" + binascii.hexlify(d).decode()


def bytesArgsToHex(args):
    for key in args:
        if type(args[key]) is bytes:
            args[key] = hexlify(args[key])
    return args


def insert_event(cur, event: logdecode.Event) -> None:
    event.args = bytesArgsToHex(event.args)
    cur.execute(
        """INSERT INTO events (transactionHash,
                                       blockNumber,
                                       address,
                                       eventName,
                                       args,
                                       blockHash,
                                       transactionIndex,
                                       logIndex,
                                       timestamp)
                               VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)""",
        (
            hexlify(event.transactionhash),
            event.blocknumber,
            event.address,
            event.name,
            json.dumps(event.args),
            hexlify(event.blockhash),
            event.transactionindex,
            event.logindex,
            event.timestamp,
        ),
    )


def insert_events(conn, events: Iterable[logdecode.Event]) -> None:
    with conn.cursor() as cur:
        for event in events:
            insert_event(cur, event)


def event_blocknumbers(events):
    """given a list of events returns the block numbers containing events"""
    return {ev.blocknumber for ev in events}


def connect(dsn):
    return psycopg2.connect(dsn, cursor_factory=psycopg2.extras.RealDictCursor)


def enrich_events(events: Iterable[logdecode.Event], blocks) -> None:
    block_by_number = {b["number"]: b for b in blocks}
    for e in events:
        block = block_by_number[e.blocknumber]
        if block["hash"] != e.blockhash:
            raise RuntimeError("bad hash! chain reorg?")
        e.timestamp = block["timestamp"]


def insert_sync_entry(conn, syncid, addresses, start_block=-1):
    """make sure we have at least one entry in the sync table"""

    with conn.cursor() as cur:
        cur.execute(
            """INSERT INTO SYNC (syncid,
                                 last_block_number,
                                 addresses,
                                 last_confirmed_block_number,
                                 latest_block_hash)
               VALUES (%s, %s, %s, %s, %s)""",
            (syncid, start_block, list(addresses), start_block, ""),
        )


def ensure_sync_entry(conn, syncid, start_block=-1):
    with conn.cursor() as cur:
        cur.execute("""select * from sync where syncid=%s""", (syncid,))
        if cur.fetchall():
            return
        cur.execute("""select addresses from sync""")
        rows = cur.fetchall()
        other_addresses = set().union(*[r["addresses"] for r in rows])

        cur.execute("""select contract_address from abis""")
        contract_addresses = set([x["contract_address"] for x in cur.fetchall()])

        addresses = contract_addresses - other_addresses
        logger.info(
            "found %s contracts, %s already being synced",
            len(contract_addresses),
            len(other_addresses),
        )
        if not addresses:
            raise RuntimeError(
                "No ABIs found. Please add some ABIs first with 'ethindex importabi'"
            )
        insert_sync_entry(conn, syncid, addresses, start_block=start_block)


def ensure_default_entry(conn, start_block=-1):
    ensure_sync_entry(conn, "default", start_block=start_block)


def delete_events(conn, fromBlock, toBlock, addresses):
    with conn.cursor() as cur:
        cur.execute(
            """DELETE FROM events
               WHERE blocknumber>=%s AND blocknumber<=%s
                     AND address in %s""",
            (fromBlock, toBlock, tuple(addresses)),
        )


class Synchronizer:
    blocks_per_round = 50000

    def __init__(
        self, conn, web3, syncid, required_confirmations=10, merge_with_syncid=None
    ):
        self.conn = conn
        self.web3 = web3
        self.syncid = syncid
        self.required_confirmations = required_confirmations
        self.merge_with_syncid = merge_with_syncid
        self.last_fully_synced_block = -1

    def _load_data_from_sync(self):
        """load the current sync status for this job from the sync table

        This also locks the row in the sync table so no two sync jobs can run
        at the same time.
        """
        with self.conn.cursor() as cur:
            cur.execute(
                """SELECT * FROM sync WHERE syncid=%s FOR UPDATE""", (self.syncid,)
            )
            row = cur.fetchone()
            self.topic_index = topic_index_from_db(
                self.conn, addresses=row["addresses"]
            )
            self.last_block_number = row["last_block_number"]
            self.last_confirmed_block_number = row["last_confirmed_block_number"]
            self.latest_block_hash = row["latest_block_hash"]

    def _sync_blocks(
        self, fromBlock, toBlock, last_confirmed_block_number, latest_block_hash
    ):
        events = get_events(self.web3, self.topic_index, fromBlock, toBlock)
        blocknumbers = event_blocknumbers(events)
        logger.info(
            "got %s events in %s out of %s blocks (%s -> %s)",
            len(events),
            len(blocknumbers),
            toBlock - fromBlock + 1,
            fromBlock,
            toBlock,
        )
        blocks = [self.web3.eth.getBlock(x) for x in blocknumbers]
        enrich_events(events, blocks)
        delete_events(self.conn, fromBlock, toBlock, self.topic_index.addresses)
        insert_events(self.conn, events)

        with self.conn.cursor() as cur:
            cur.execute(
                """UPDATE sync
                   SET last_block_number=%s, last_confirmed_block_number=%s, latest_block_hash=%s
                   WHERE syncid=%s""",
                (toBlock, last_confirmed_block_number, latest_block_hash, self.syncid),
            )

    def _try_merge(self, cur):
        cur.execute(
            """SELECT * FROM sync WHERE syncid in %s FOR UPDATE""",
            ((self.merge_with_syncid, self.syncid),),
        )
        rows = cur.fetchall()

        assert len(rows) == 2
        if rows[0]["syncid"] == self.merge_with_syncid:
            dst, src = rows
        else:
            src, dst = rows

        block_diff = dst["last_block_number"] - src["last_block_number"]

        if block_diff == 0:
            if dst["latest_block_hash"] != src["latest_block_hash"]:
                logger.info(
                    "cannot merge runsync jobs, because they see a different state of the chain"
                )
                return False

            logger.info(
                "merging sync job %s into %s", self.syncid, self.merge_with_syncid
            )
            cur.execute(
                """UPDATE sync
                           SET addresses=%s
                           WHERE syncid=%s""",
                (dst["addresses"] + src["addresses"], self.merge_with_syncid),
            )
            cur.execute("delete from sync where syncid=%s", (self.syncid,))
            return True
        elif block_diff < 0:
            logger.info(
                "cannot merge runsync  jobs, we are %s blocks in front of %s",
                -block_diff,
                self.merge_with_syncid,
            )
        else:
            logger.info(
                "cannot merge runsync jobs, we are %s blocks behind %s",
                block_diff,
                self.merge_with_syncid,
            )
        return False

    def try_merge(self):
        with self.conn:
            with self.conn.cursor() as cur:
                return self._try_merge(cur)

    def sync_round(self):
        self._load_data_from_sync()
        latest_block = self.web3.eth.getBlock("latest")
        latest_block_hash = hexlify(latest_block["hash"])
        latest_block_number = latest_block["number"]
        fromBlock = self.last_confirmed_block_number + 1
        toBlock = min(
            latest_block_number,
            self.last_confirmed_block_number + self.blocks_per_round,
        )
        last_confirmed_block_number = max(
            min(toBlock, latest_block_number - self.required_confirmations), -1
        )
        if fromBlock <= toBlock and (
            self.last_block_number != latest_block_number
            or self.latest_block_hash != latest_block_hash
        ):
            self._sync_blocks(
                fromBlock, toBlock, last_confirmed_block_number, latest_block_hash
            )
            finished = False
        else:
            if self.last_fully_synced_block != toBlock:
                self.last_fully_synced_block = toBlock
                logger.info("already synced up to latest block %s", toBlock)
            finished = True

        self.conn.commit()
        return finished

    def sync_loop(self, waittime):
        while 1:
            self.sync_until_current()
            if self.merge_with_syncid and self.try_merge():
                return

            time.sleep(waittime)

    def sync_until_current(self):
        while not self.sync_round():
            pass


@click.command()
@click.option("--jsonrpc", help="jsonrpc URL to use", default="http://127.0.0.1:8545")
@click.option(
    "--required-confirmations",
    help="number of confirmations until we consider a block final",
    default=10,
)
@click.option(
    "--waittime",
    help="time to sleep in milliseconds waiting for a new block",
    default=1000,
)
@click.option(
    "--startblock", help="Block from where events should be synced", default=-1
)
@click.option("--syncid", help="syncid to use", default="default")
@click.option("--merge-with-syncid", help="syncid to merge with")
def runsync(
    jsonrpc, waittime, startblock, required_confirmations, syncid, merge_with_syncid
):
    logging.basicConfig(level=logging.INFO)
    logger.info("version %s starting", util.get_version())

    # we like to survive a postgresql restart, so we need to catch errors here,
    # since we must create a new connection in that case.
    while 1:
        try:
            web3 = Web3(Web3.HTTPProvider(jsonrpc, request_kwargs={"timeout": 60}))
            with connect("") as conn:
                ensure_sync_entry(conn, syncid)
                s = Synchronizer(
                    conn,
                    web3,
                    syncid,
                    required_confirmations=required_confirmations,
                    merge_with_syncid=merge_with_syncid,
                )
                s.sync_loop(waittime * 0.001)
                break
        except Exception:
            logger.error(
                "An error occured in runsync. Will restart runsync in 10 seconds",
                exc_info=sys.exc_info(),
            )
            time.sleep(10)


def do_importabi(conn, addresses, contracts):
    a2abi = logdecode.build_address_to_abi_dict(
        json.load(open(addresses)), json.load(open(contracts))
    )
    logger.info("importing %s abis", len(a2abi))
    with conn:
        with conn.cursor() as cur:
            for contract_address, abi in a2abi.items():
                cur.execute(
                    """INSERT INTO abis (contract_address, abi)
                       VALUES (%s, %s)
                       ON CONFLICT(contract_address) DO NOTHING""",
                    (contract_address, json.dumps(abi)),
                )


@click.command()
@click.option(
    "--addresses",
    default="./addresses.json",
    show_default=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "--contracts",
    default="./contracts.json",
    show_default=True,
    type=click.Path(exists=True, dir_okay=False),
)
def importabi(addresses, contracts):
    logging.basicConfig(level=logging.INFO)
    logger.info("version %s starting", util.get_version())
    do_importabi(connect(""), addresses, contracts)


def do_createtables(conn):
    with conn:
        with conn.cursor() as cur:
            cur.execute(
                """
                CREATE TABLE events (
                     transactionHash TEXT NOT NULL,
                     blockNumber INTEGER NOT NULL,
                     address TEXT NOT NULL,
                     eventName TEXT NOT NULL,
                     args JSONB,
                     blockHash TEXT NOT NULL,
                     transactionIndex INTEGER NOT NULL,
                     logIndex INTEGER NOT NULL,
                     timestamp INTEGER NOT NULL,
                     PRIMARY KEY(transactionHash, address, blockHash, transactionIndex, logIndex)
                   );

                  CREATE TABLE sync (
                    syncid TEXT NOT NULL PRIMARY KEY,
                    last_block_number INTEGER NOT NULL,
                    addresses TEXT[] NOT NULL,
                    last_confirmed_block_number INTEGER NOT NULL,
                    latest_block_hash TEXT NOT NULL
                  );

                  CREATE TABLE abis (
                    contract_address TEXT NOT NULL PRIMARY KEY,
                    abi JSONB NOT NULL
                  );"""
            )


@click.command()
def createtables():
    logging.basicConfig(level=logging.INFO)
    logger.info("version %s starting", util.get_version())
    logger.info("creating tables")
    do_createtables(connect(""))


def do_droptables(conn, force):
    with conn:
        with conn.cursor() as cur:
            for table in ["events", "sync", "abis"]:
                stmt = "DROP TABLE IF EXISTS {}".format(table)
                logger.info("executing %r", stmt)
                if force:
                    cur.execute(stmt)


@click.command()
@click.option("--force", help="really delete the tables", is_flag=True)
def droptables(force):
    """drop database tables"""
    logging.basicConfig(level=logging.INFO)
    logger.info("version %s starting", util.get_version())
    if not force:
        logger.warn("dry-run, please specify --force to really delete the tables")

    do_droptables(connect(""), force)

    sys.exit(0 if force else 1)  # just in case we forget to add --force



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="eth-index",
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    description="Indexer for the ethereum blockchain",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/trustlines-protocol/py-eth-index",
    # Author details
    author="Trustlines-Network",
    author_email="contact@brainbot.com",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 2 - Pre-Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    # What does your project relate to?
    keywords="ethereum indexer database",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["web3>=4.4.1", "psycopg2>=2.7,<2.8", "click", "attrs"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={},
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={},
    python_requires=">=3.6",
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],
    entry_points="""
    [console_scripts]
    ethindex=ethindex.cli:cli
    """,
)
